@@ -3,6 +3,9 @@ from enum import Enum
 from functools import partial
 import math
 import os
+from queue import Queue
+import threading
+import time
 
 import requests
 
@@ -11,6 +14,30 @@ from ..metadata import MetadataItem, to_api_format
 from .. import utils
 
 
+INVALID_FILENAME_CHARACTERS = {"\\", "/", ":"}
+
+
+class ThreadCounter():
+    """ Thread safe counter"""
+
+    def __init__(self):
+        self._count = 0
+        self.lock = threading.Lock()
+
+    @property
+    def count(self):
+        with self.lock:
+            return self._count
+
+    def add(self):
+        with self.lock:
+            self._count += 1
+
+    def subtract(self):
+        with self.lock:
+            self._count -= 1
+
+
 class FileTransferProgressHandler:
     """
     Invokes a callback when transfer size thresholds have been reached.
@@ -209,7 +236,8 @@ class FilesAPI(GenericAPI):
 
         return output
 
-    def get_all_from_location(self, space_id, path=None, file_type=None):
+    def get_all_from_location(self, space_id, path=None, file_type=None,
+                              recurse=False):
         """
         Get all files/folders in a given file system location.
 
@@ -224,12 +252,72 @@ class FilesAPI(GenericAPI):
         file_type : hyperthought.api.files.FilesAPI.FileType or None
             An enum instance specifying the type of files to get, i.e. files,
             folders, or both.  Will get both if None.
+        recurse : bool
+            A boolean specifying whether to recursively get from child folders.
+            Will default to False if omimtted.
 
         Returns
         -------
         A list of dicts containing information on files and/or folders in a
         given file system location.
         """
+
+        def _get_all_from_location_flat(space_id, path, file_type=None):
+            """Get all top-level files/folders from a specific location."""
+            # Set pagination parameters.
+            start = 0
+            length = 25
+
+            # Create list to collect output.
+            files = []
+
+            while True:
+                new_files = self.get_from_location(
+                    space_id=space_id,
+                    path=path,
+                    file_type=file_type,
+                    start=start,
+                    length=length,
+                )
+
+                if not new_files:
+                    break
+
+                files.extend(new_files)
+                start += length
+
+            return files
+
+        def _queue_folder(folder):
+            """Create partial functions to get all files within a folder."""
+            # Set pagination parameters.
+            start = 0
+            length = 25
+
+            while start < folder['items']:
+                get_queue.put(
+                    partial(
+                        self.get_from_location,
+                        space_id=space_id,
+                        path=folder['path'] + folder['pk'] + ',',
+                        start=start,
+                        length=length
+                    )
+                )
+                start += length
+
+        def _run_request(get_request):
+            """Run get file request and queue any folders."""
+            files = get_request()
+
+            for file in files:
+                if file['ftype'] == 'Folder':
+                    _queue_folder(file)
+
+            files_recursive.extend(files)
+
+            active_threads.subtract()
+
         # Validate inputs.
 
         if not isinstance(space_id, str):
@@ -244,29 +332,70 @@ class FilesAPI(GenericAPI):
                 "or None."
             )
 
-        # Set pagination parameters.
-        start = 0
-        length = 25
-
-        # Create list to collect output.
-        files = []
-
-        while True:
-            new_files = self.get_from_location(
-                space_id=space_id,
-                path=path,
-                file_type=file_type,
-                start=start,
-                length=length,
+        if not isinstance(recurse, bool):
+            raise ValueError(
+                "The recurse parameter to get_all_from_location must be a "
+                "boolean."
             )
 
-            if not new_files:
-                break
+        if recurse is False:
 
-            files.extend(new_files)
-            start += length
+            files = _get_all_from_location_flat(
+                space_id=space_id,
+                path=path,
+                file_type=file_type
+            )
 
-        return files
+            return files
+
+        else:
+
+            print(
+                "Getting all child items recursively, this may take "
+                "several minutes..."
+            )
+
+            MAX_THREADS = 100
+            get_queue = Queue()
+            active_threads = ThreadCounter()
+            files_recursive = _get_all_from_location_flat(
+                space_id=space_id,
+                path=path,
+            )
+
+            for file in files_recursive:
+                if file['ftype'] == 'Folder':
+                    _queue_folder(file)
+
+            while not get_queue.empty() or active_threads.count > 0:
+
+                if get_queue.empty() or active_threads.count >= MAX_THREADS:
+                    time.sleep(1)
+
+                else:
+                    active_threads.add()
+                    get_request = get_queue.get()
+                    thread = threading.Thread(
+                        target=_run_request,
+                        args=[get_request],
+                        daemon=True
+                    )
+                    thread.start()
+
+            if file_type == self.FileType.FOLDERS_ONLY:
+                return [
+                    response_item for response_item in files_recursive if
+                    response_item["ftype"] == "Folder"
+                ]
+
+            elif file_type == self.FileType.FILES_ONLY:
+                return [
+                    response_item for response_item in files_recursive if
+                    response_item["ftype"] != "Folder"
+                ]
+
+            else:
+                return files_recursive
 
     def get_id(self, name, space_id, path=None):
         """
@@ -809,6 +938,37 @@ class FilesAPI(GenericAPI):
             # NOTE: This method throws exceptions.
             self._report_api_error(r)
 
+    def update_name(self, file_id, new_name):
+        """
+        Update name for a file.
+
+        Parameters
+        ----------
+        file_id : str
+            The id (uuid) for the file of interest.
+        new_metadata : str
+            New name for the file.
+        """
+        # NOTE:  This method throws exceptions.
+        new_name = self._validate_new_name(new_name)
+        curried_request = partial(
+            requests.patch,
+            url=f"{self._base_url}/api/files/",
+            json={
+                'file_id': file_id,
+                'updates': {
+                    'content': {
+                        'name': new_name
+                    }
+                }
+            },
+        )
+        r = self.attempt_api_call(curried_request=curried_request)
+
+        if r.status_code >= ERROR_THRESHOLD:
+            # NOTE:  This method throws exceptions.
+            self._report_api_error(r)
+
     def upload_using_url(self, upload_url, local_path,
                          progress_callback=None, n_chunks=100):
         """
@@ -881,7 +1041,7 @@ class FilesAPI(GenericAPI):
 
         file_handle.close()
 
-    def generate_blank_upload_urls(self, url_count):
+    def generate_blank_upload_urls(self, url_count, failed_url=None):
         """
         Generate blank upload urls.
 
@@ -891,6 +1051,11 @@ class FilesAPI(GenericAPI):
         ----------
         url_count : int
             The number of urls to generate.
+        failedUrl : str or None
+            A previously requested url that didn't work.
+            This will be used to exclude a server from consideration
+            for url generation -- if there are multiple servers associated
+            with the files backend.
 
         Returns
         -------
@@ -905,7 +1070,8 @@ class FilesAPI(GenericAPI):
             requests.post,
             url=f"{self._files_base_url}generate-blank-upload-urls/",
             json={
-                'urlCount': url_count
+                'urlCount': url_count,
+                "failedUrl": failed_url,
             },
         )
         r = self.attempt_api_call(curried_request=curried_request)
@@ -1220,6 +1386,12 @@ class FilesAPI(GenericAPI):
         assert isinstance(name, str)
         return name
 
+    def _validate_new_name(self, name):
+        name = self._validate_name(name)
+        for char in name:
+            assert char not in INVALID_FILENAME_CHARACTERS
+        return name
+
     def _validate_size(self, size):
         assert isinstance(size, int) and size >= 0
         return size
