@@ -2,6 +2,7 @@ import collections
 from datetime import datetime
 from functools import partial
 import uuid
+from xml.etree import ElementTree
 from xml.etree.ElementTree import tostring
 
 import requests
@@ -1333,3 +1334,124 @@ class WorkflowAPI(GenericAPI):
             self._report_api_error(response=r)
 
         return r.json()
+
+    def _remove_from_siblings(self, element):
+        """
+        Remove predecessor/successor relationships.
+
+        Called from delete.
+
+        Parameters
+        ----------
+        element : dict
+            Document for a workflow element.
+        """
+        def delete_link(source_id, target_id):
+            """Delete a predecessor->successor relationship."""
+            url = (
+                f"{self._base_url}/api/workflow/process/"
+                f"{source_id}/link/{target_id}/"
+            )
+            curried_request = partial(requests.delete, url=url)
+            r = self.attempt_api_call(curried_request=curried_request)
+
+            if r.status_code >= ERROR_THRESHOLD:
+                # NOTE:  This method will throw an exception.
+                self._report_api_error(response=r)
+
+        element_id = element["id"]
+
+        for predecessor_id in element["predecessors"]:
+            delete_link(source_id=predecessor_id, target_id=element_id)
+
+        for successor_id in element["successors"]:
+            delete_link(source_id=element_id, target_id=successor_id)
+
+    def _remove_from_parent(self, element):
+        """
+        Remove a workflow element from its parent workflow.
+
+        Called from delete.
+
+        Parameters
+        ----------
+        element : dict
+            Document for a workflow element.
+        """
+        element_id = element["id"]
+        parent_id = element["parentProcess"]
+
+        if not parent_id:
+            return
+
+        parent = self.get_document(document_id=parent_id)
+
+        # Remove id from children list.
+        child_ids = parent["children"]
+        parent["children"] = [
+            child_id
+            for child_id in child_ids
+            if child_id != element_id
+        ]
+
+        # Modify XML as needed.
+        # (Otherwise, leave XML as is.  Don't regenerate.)
+        tree = ElementTree.fromstring(parent["xml"])
+        root_node = tree.find(".//root")
+        client_id = element["clientId"]
+        subelements_to_remove = []
+        attribute_names = {"id", "source", "target"}
+
+        for subelement in root_node:
+            for attribute_name, value in subelement.attrib.items():
+                if attribute_name in attribute_names and value == client_id:
+                    subelements_to_remove.append(subelement)
+
+        for subelement in subelements_to_remove:
+            root_node.remove(subelement)
+
+        def xml_to_string(xml_):
+            return ElementTree.tostring(
+                xml_,
+                encoding="utf8",
+                method="xml",
+            ).decode()
+
+        xml_string = xml_to_string(tree)
+        xml_string = xml_string.replace(
+            "<?xml version='1.0' encoding='utf8'?>", "").strip()
+        parent["xml"] = xml_string
+        self.update_document(updated_document=parent)
+
+    def _delete_element(self, element_id):
+        """
+        Delete a workflow element itself by calling the appropriate endpoint.
+
+        Called from delete.
+
+        Parameters
+        ----------
+        element_id : str
+            The id for a workflow element.
+        """
+        url = f"{self._base_url}/api/workflow/process/{element_id}/"
+        curried_request = partial(requests.delete, url=url)
+        r = self.attempt_api_call(curried_request=curried_request)
+
+        if r.status_code >= ERROR_THRESHOLD:
+            # NOTE:  This method will throw an exception.
+            self._report_api_error(response=r)
+
+    def delete(self, element_id):
+        """
+        Delete a workflow element.
+
+        Parameters
+        ----------
+        element_id : str
+            The id for a workflow element.
+        """
+        element = self.get_document(document_id=element_id)
+        self._remove_from_siblings(element=element)
+        self._remove_from_parent(element=element)
+        self._delete_element(element_id=element_id)
