@@ -43,6 +43,15 @@ class AuthExpirationException(Exception):
     pass
 
 
+class ManualTokenRefreshNotSpecifiedException(Exception):
+    """
+    To be called when a user attempts to manually refresh a token
+    on a TokenAuthentication object for which manual refresh has not been
+    specified.
+    """
+    pass
+
+
 class Authentication(abc.ABC):
     """
     Abstract base class for authentication classes.  Defines common interface.
@@ -78,6 +87,10 @@ class TokenAuthentication(Authentication):
         Base-64 encoded text containing a dict with authorization info.
     verify : bool
         SSL verification, set to True by default.
+    manual_refresh : bool
+        If True, the client will be responsible for refreshing the token
+        using the `refresh` method.  This must be done before the token
+        expires.
     delayed_refresh : bool
         This parameter is mainly useful for unit tests.  Accept the default
         (False) otherwise.
@@ -121,9 +134,16 @@ class TokenAuthentication(Authentication):
         'refreshToken',     # Token to refresh access / get a new access token.
     )
 
-    def __init__(self, auth_payload, verify=True, delayed_refresh=False):
+    def __init__(
+        self,
+        auth_payload,
+        verify=True,
+        manual_refresh=False,
+        delayed_refresh=False,
+    ):
         super().__init__(verify)
         self._delayed_refresh = bool(delayed_refresh)
+        self._manual_refresh = bool(manual_refresh)
         auth_payload = json.loads(base64.b64decode(auth_payload))
 
         # TODO:  Improve upon parameter validation.
@@ -156,10 +176,43 @@ class TokenAuthentication(Authentication):
         time_delta = self._expiration_time - current_time
         self._seconds_to_expiration = time_delta.total_seconds()
 
-        # Define a refresh thread without starting it.
-        self._refresh_thread = Thread(target=self._refresh)
-        self._refresh_thread.setDaemon(True)
-        self._refresh_thread.start()
+        # Start a thread to refresh the token automatically, as needed.
+        if not self._manual_refresh:
+            self._refresh_thread = Thread(target=self._auto_refresh)
+            self._refresh_thread.setDaemon(True)
+            self._refresh_thread.start()
+
+    @property
+    def expiration_time(self):
+        return self._expiration_time
+
+    @property
+    def seconds_to_expiration(self):
+        return self._seconds_to_expiration
+
+    @property
+    def access_token(self):
+        return self._access_token
+
+    @access_token.setter
+    def access_token(self, value):
+        if not self._manual_refresh:
+            raise ManualTokenRefreshNotSpecifiedException(
+                "Manual reset not specified.  Reset not allowed.")
+
+        self._access_token = value
+
+    @property
+    def refresh_token(self):
+        return self._refresh_token
+
+    @refresh_token.setter
+    def refresh_token(self, value):
+        if not self._manual_refresh:
+            raise ManualTokenRefreshNotSpecifiedException(
+                "Manual refresh not specified.  Reset not allowed.")
+
+        self._refresh_token = value
 
     def get_base_url(self):
         """Get the base url for the auth client."""
@@ -230,9 +283,68 @@ class TokenAuthentication(Authentication):
 
         return self._user_info['email']
 
-    def _refresh(self):
+    def refresh(self, max_attempts=15):
         """
-        Refresh the access and refresh tokens using the refresh token.
+        Manually refresh auth token.
+
+        This function will only work if it is called before the current token
+        expires, and manual token refresh has been specified in the constructor\
+        parameters.
+        """
+        if not self._manual_refresh:
+            raise ManualTokenRefreshNotSpecifiedException(
+                "Manual refresh not specified.")
+
+        refreshed = False
+        attempts = 0
+
+        while not refreshed and attempts < max_attempts:
+            attempts += 1
+
+            if self._seconds_to_expiration < 0:
+                raise AuthExpirationException(
+                    "Unable to authenticate with expired token")
+
+            # Request object, defined here for post-loop error handling.
+            r = None
+
+            data = {
+                'client_id': self._client_id,
+                'client_secret': self._client_secret,
+                'grant_type': 'refresh_token',
+                'refresh_token': self._refresh_token,
+            }
+
+            r = requests.post(
+                url=self._token_url,
+                data=data,
+                cookies=self._cookies,
+                verify=self.verify,
+            )
+
+            if r.status_code >= 400:
+                time.sleep(REFRESH_RETRY_SLEEP_TIME)
+                continue
+
+            auth_info = r.json()
+            self._access_token = auth_info['access_token']
+            self._refresh_token = auth_info['refresh_token']
+            self._seconds_to_expiration = auth_info['expires_in']
+            self._expiration_time = datetime.now(tzlocal()) + timedelta(
+                seconds=self._seconds_to_expiration)
+            refreshed = True
+
+        if not refreshed:
+            raise AuthExpirationException(
+                "Unable to refresh token. "
+                f"Call to {self._token_url} has failed. "
+                f"Status code: {r.status_code}. "
+                f"Response: {r.content.decode()}."
+            )
+
+    def _auto_refresh(self):
+        """
+        Automatically refresh the token.
 
         This method is the target of a separate execution thread.
         It should run as long as the client application is running, to ensure
@@ -335,13 +447,13 @@ class ClientAuthentication(Authentication):
 
     def __init__(self, client_id, client_secret, auth_base_url, verify=True):
         super().__init__(verify)
-        self.AUTH_POST_DATA = "grant_type=client_credentials"
-        self.AUTH_CONTENT_TYPE = "application/x-www-form-urlencoded"
+        self._auth_post_data = "grant_type=client_credentials"
+        self._auth_content_type = "application/x-www-form-urlencoded"
 
-        self.client_id = client_id
-        self.client_secret = client_secret
-        self.auth_base_url = auth_base_url
-        self.auth_target_url = f"{auth_base_url}/openid/token/"
+        self._client_id = client_id
+        self._client_secret = client_secret
+        self._auth_base_url = auth_base_url
+        self._auth_target_url = f"{auth_base_url}/openid/token/"
 
         # This cookie is needed to avoid being redirected to the HyperThought
         # banner page.
@@ -349,11 +461,25 @@ class ClientAuthentication(Authentication):
             'dodAccessBanner': 'true'
         }
 
-        # This data members will be set by self._update_auth,
+        # These data members will be set by self._update_auth,
         # which will be called as needed by self.get_headers.
         self._access_token = None
         self._expiration = None
 
+        self._user_info = None
+
+    @property
+    def client_id(self):
+        return self._client_id
+
+    @property
+    def client_secret(self):
+        return self._client_secret
+
+    @property
+    def auth_base_url(self):
+        return self._auth_base_url
+
     def get_headers(self):
         """Get headers for authenticated REST API requests."""
 
@@ -370,19 +496,19 @@ class ClientAuthentication(Authentication):
 
     def get_base_url(self):
         """Get the base url for the auth client."""
-        return self.auth_base_url
+        return self._auth_base_url
 
     def get_base_url_dns(self):
         """Get the base url DNS for the auth client."""
-        if '//' in self.auth_base_url:
-            split_dns_name = self.auth_base_url.split('//')[1].split(':')
+        if '//' in self._auth_base_url:
+            split_dns_name = self._auth_base_url.split('//')[1].split(':')
 
             if len(split_dns_name) > 1:
                 return '.'.join(split_dns_name)
 
             return split_dns_name[0]
         else:
-            return self.auth_base_url
+            return self._auth_base_url
 
     def _update_auth(self):
         """
@@ -391,14 +517,14 @@ class ClientAuthentication(Authentication):
         The bootstrap request will yield
         """
         r = requests.post(
-            url=self.auth_target_url,
-            data=self.AUTH_POST_DATA,
+            url=self._auth_target_url,
+            data=self._auth_post_data,
             headers={
-                'Content-Type': self.AUTH_CONTENT_TYPE,
+                'Content-Type': self._auth_content_type,
             },
             auth=requests.auth.HTTPBasicAuth(
-                self.client_id,
-                self.client_secret,
+                self._client_id,
+                self._client_secret,
             ),
             cookies=self.get_cookies(),
             verify=self.verify,
@@ -408,6 +534,7 @@ class ClientAuthentication(Authentication):
         # The expired token will result in an API error when used.
         # Retrying the request will result in another invocation of
         # get_headers, thus another invocation of this method.
+        # TODO:  Log error rather than printing it.
         if r.status_code >= 300:
             print(
                 f"Authentication failure. Status code: {r.status_code}. "
@@ -429,3 +556,46 @@ class ClientAuthentication(Authentication):
             or
             self._expiration <= datetime.now(tzlocal())
         )
+
+    def _set_user_info(self):
+        """Set self._user_info by calling the userinfo endpoint."""
+        url = f"{self.auth_base_url.rstrip('/')}/api/auth/userinfo/"
+        r = requests.get(
+            url=url,
+            headers=self.get_headers(),
+            cookies=self.get_cookies(),
+            verify=self.verify,
+        )
+
+        if r.status_code >= 400:
+            raise Exception('Unable to set user info.')
+
+        self._user_info = r.json()
+
+    def get_username(self):
+        """Get the username associated with the currently logged in user."""
+        if not self._user_info:
+            self._set_user_info()
+
+        return self._user_info['username']
+
+    def get_first_name(self):
+        """Get the first name of the currently logged-in user."""
+        if not self._user_info:
+            self._set_user_info()
+
+        return self._user_info['first_name']
+
+    def get_last_name(self):
+        """Get the last name of the currently logged-in user."""
+        if not self._user_info:
+            self._set_user_info()
+
+        return self._user_info['last_name']
+
+    def get_email(self):
+        """Get the email address of the currently logged-in user."""
+        if not self._user_info:
+            self._set_user_info()
+
+        return self._user_info['email']
